prompts = {
    "check_end_condition": {
        "question": """Please determine if the task is completed based on the ontology generated from the user's input text. Respond with 'done' or 'not_done'.""",
        "additional": """Current ontology graph: """
    },
    "generate_nodes": {
        "question": """Please identify and list the relationships between nodes for the ontology based on the user's input text: """,
        "additional": """Possible relationships include: """
    },
}


import os
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_text_splitters import CharacterTextSplitter
from langchain_community.vectorstores import FAISS
from langchain_huggingface import HuggingFaceEmbeddings
from langchain_core.runnables import RunnablePassthrough
from langchain_core.pydantic_v1 import BaseModel, Field
from typing import List
from langchain_community.document_loaders import TextLoader
from .relations import RELATIONS
from .prompts import prompts


CACHE_FILE = ".cache"


def initialize_text_chain(text, parser, language_model, prompt_template, embedding_model=HuggingFaceEmbeddings(),
                          embedding_chunk_size=500):
    """Initialize the text processing chain with caching and text splitting."""
    manage_cache_file(create=True, content=text)

    loader = TextLoader(CACHE_FILE)
    documents = loader.load()
    splitter = CharacterTextSplitter(chunk_size=embedding_chunk_size, chunk_overlap=0)
    texts = splitter.split_documents(documents)

    embedding = embedding_model
    vector_store = FAISS.from_documents(texts, embedding)
    retriever = vector_store.as_retriever()

    return ({"context": retriever, "query": RunnablePassthrough()} | prompt_template | language_model | parser)


def manage_cache_file(create=False, content=None):
    """Handle cache file creation and removal."""
    if create:
        if os.path.exists(CACHE_FILE):
            os.remove(CACHE_FILE)
        with open(CACHE_FILE, 'w') as file:
            file.write(content if content else "")


def check_termination_condition(input_text, result_graph, processing_chain, attempt_count, max_attempts):
    """Check if the processing should terminate based on conditions."""
    if attempt_count >= max_attempts:
        return True

    graph_as_string = str(result_graph)
    result = processing_chain.invoke(
        prompts['check_end_condition']['question'] + prompts['check_end_condition']['additional'] + graph_as_string
    )

    try:
        if result.completion == "done":
            return True
    except Exception:
        return False
    return False


def split_text_into_chunks(text, chunk_size=50):
    """Split text into manageable chunks."""
    words = text.split()
    return [' '.join(words[i:i + chunk_size]) for i in range(0, len(words), chunk_size)]


def process_text_chunks(chain, text, result_graph, relations, process_chunk_size):
    """Process individual text chunks to generate nodes."""
    chunks = split_text_into_chunks(text, process_chunk_size)
    for chunk in chunks:
        try:
            result = chain.invoke(
                prompts['generate_nodes']['question'] + chunk + prompts['generate_nodes']['additional'] + str(relations)
            )
            for item in result.relations:
                result_graph.append({
                    "from_node": item.from_node,
                    "relation": item.relation,
                    "to_node": item.to_node
                })
        except Exception:
            continue
    return result_graph


def define_relation_model(relations):
    """Define Pydantic models for relations."""

    class RelationModel(BaseModel):
        from_node: str = Field(..., description="Source node of the relation.")
        to_node: str = Field(..., description="Target node of the relation.")
        relation: str = Field(..., description="Type of relation.")

    class RelationsModel(BaseModel):
        relations: List[RelationModel]

    return RelationsModel


def define_task_completion_model():
    """Define Pydantic model for task completion."""

    class TaskCompletionModel(BaseModel):
        completion: str = Field(..., description="Completion status of the task.")

    return TaskCompletionModel


def prepare_parser_and_prompt(model):
    """Prepare parser and prompt template for structured outputs."""
    parser = PydanticOutputParser(pydantic_object=model)
    prompt_template = ChatPromptTemplate.from_messages([
        ("system",
         "Answer the user query. Wrap the output in `json` tags\n{format_instructions}. The context is {context}"),
        ("human", "{query}"),
    ]).partial(format_instructions=parser.get_format_instructions())

    return parser, prompt_template


def execute_graph_generation(text, language_model, embedding_model=HuggingFaceEmbeddings(), relations=RELATIONS,
                             max_attempts=10, process_chunk_size=50, embedding_chunk_size=500):
    """Generate the result graph through iterative node generation and condition checks."""
    RelationsModel = define_relation_model(relations)
    TaskCompletionModel = define_task_completion_model()

    parser, prompt_template = prepare_parser_and_prompt(RelationsModel)
    initial_chain = initialize_text_chain(text, parser, language_model, prompt_template, embedding_model, embedding_chunk_size)

    result_graph = []
    finished = False
    attempts = 0

    while not finished:
        attempts += 1
        result_graph = process_text_chunks(initial_chain, text, result_graph, relations, process_chunk_size)
        graph_as_string = str(result_graph)
        parser, prompt_template = prepare_parser_and_prompt(TaskCompletionModel)
        completion_chain = initialize_text_chain(graph_as_string, parser, language_model, prompt_template,
                                                 embedding_model, embedding_chunk_size)
        finished = check_termination_condition(text, result_graph, completion_chain, attempts, max_attempts)

    manage_cache_file()
    return result_graph


from setuptools import setup, find_packages


setup(
    name='eknowledge',
    version='0.1.4',
    author='Eugene Evstafev',
    author_email='chigwel@gmail.com',
    description='A Python package for executing graph generation from textual inputs.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/chigwell/eknowledge',
    packages=find_packages(),
    install_requires=[
        'faiss-cpu==1.8.0.post1',
        'huggingface-hub==0.23.4',
        'langchain==0.2.6',
        'langchain-community==0.2.6',
        'langchain-core==0.2.10',
        'langchain-huggingface==0.0.3',
        'langchain-text-splitters==0.2.2',
        'langsmith==0.1.82',
        'packaging==24.1',
        'pydantic==2.7.4',
        'pydantic_core==2.18.4',
        'safetensors==0.4.3',
        'scikit-learn==1.5.0',
        'scipy==1.14.0',
        'sentence-transformers==3.0.1',
        'torch==2.3.1',
        'transformers==4.42.3',
        'typing_extensions==4.12.2'
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)


import unittest
from unittest.mock import patch, MagicMock
from eknowledge.main import (
    initialize_text_chain, check_termination_condition, split_text_into_chunks,
    process_text_chunks, execute_graph_generation
)
from eknowledge.relations import RELATIONS

class TestEKnowledge(unittest.TestCase):
    def test_initialize_text_chain_creates_cache_file(self):
        with patch('eknowledge.main.os.path.exists') as mock_exists, \
             patch('eknowledge.main.open', unittest.mock.mock_open()) as mock_file, \
             patch('eknowledge.main.TextLoader') as MockTextLoader, \
             patch('eknowledge.main.CharacterTextSplitter') as MockTextSplitter, \
             patch('eknowledge.main.FAISS') as MockFAISS:
            mock_exists.return_value = False
            MockTextLoader.return_value.load.return_value = ['test text']
            MockTextSplitter.return_value.split_documents.return_value = ['test text']
            MockFAISS.from_documents.return_value.as_retriever.return_value = 'retriever'

            chain = initialize_text_chain('test text', MagicMock(), MagicMock(), MagicMock())
            mock_file.assert_called_with('.cache', 'w')
            self.assertIsNotNone(chain)

    def test_check_termination_condition_max_attempts_exceeded(self):
        result = check_termination_condition('input', 'result_graph', MagicMock(), 5, 5)
        self.assertTrue(result)

    def test_split_text_into_chunks(self):
        chunks = split_text_into_chunks("this is a test text for splitting", 3)
        expected_chunks = ["this is a", "test text for", "splitting"]
        self.assertEqual(chunks, expected_chunks)

    def test_process_text_chunks_handles_exceptions(self):
        with patch('eknowledge.main.split_text_into_chunks') as mock_split:
            mock_split.return_value = ["chunk1", "chunk2"]
            chain = MagicMock()
            chain.invoke.side_effect = [Exception("failed"), MagicMock()]

            result_graph = process_text_chunks(chain, "input text", [], RELATIONS)
            self.assertEqual(len(result_graph), 0)

    def test_execute_graph_generation_integration(self):
        with patch('eknowledge.main.initialize_text_chain') as mock_initialize_text_chain, \
             patch('eknowledge.main.check_termination_condition') as mock_check_termination_condition, \
             patch('eknowledge.main.process_text_chunks') as mock_process_text_chunks:
            mock_initialize_text_chain.side_effect = [MagicMock(), MagicMock()]
            mock_check_termination_condition.return_value = True
            mock_process_text_chunks.return_value = []

            graph = execute_graph_generation("input text", MagicMock())
            self.assertIsInstance(graph, list)
            mock_check_termination_condition.assert_called()

if __name__ == '__main__':
    unittest.main()